"""Defines the CLI for getting information about robot classes."""

import itertools
import json
import logging
import math
import time
from typing import Sequence

import click
from tabulate import tabulate

from kscale.utils.cli import coro
from kscale.web.clients.robot_class import RobotClassClient
from kscale.web.gen.api import RobotURDFMetadataInput

logger = logging.getLogger(__name__)


class RobotURDFMetadataInputStrict(RobotURDFMetadataInput):
    class Config:
        extra = "forbid"


@click.group()
def cli() -> None:
    """Get information about robot classes."""
    pass


@cli.command()
@coro
async def list() -> None:
    """Lists all robot classes."""
    client = RobotClassClient()
    robot_classes = await client.get_robot_classes()
    if robot_classes:
        # Prepare table data
        table_data = [
            [
                click.style(rc.id, fg="blue"),
                click.style(rc.class_name, fg="green"),
                rc.description or "N/A",
            ]
            for rc in robot_classes
        ]
        click.echo(tabulate(table_data, headers=["ID", "Name", "Description"], tablefmt="simple"))
    else:
        click.echo(click.style("No robot classes found", fg="red"))


@cli.command()
@click.argument("name")
@click.option("-d", "--description", type=str, default=None)
@coro
async def add(
    name: str,
    description: str | None = None,
) -> None:
    """Adds a new robot class."""
    async with RobotClassClient() as client:
        robot_class = await client.create_robot_class(name, description)
    click.echo("Robot class created:")
    click.echo(f"  ID: {click.style(robot_class.id, fg='blue')}")
    click.echo(f"  Name: {click.style(robot_class.class_name, fg='green')}")
    click.echo(f"  Description: {click.style(robot_class.description or 'N/A', fg='yellow')}")


@cli.command()
@click.argument("current_name")
@click.option("-n", "--name", type=str, default=None)
@click.option("-d", "--description", type=str, default=None)
@coro
async def update(current_name: str, name: str | None = None, description: str | None = None) -> None:
    """Updates a robot class."""
    async with RobotClassClient() as client:
        robot_class = await client.update_robot_class(current_name, name, description)
    click.echo("Robot class updated:")
    click.echo(f"  ID: {click.style(robot_class.id, fg='blue')}")
    click.echo(f"  Name: {click.style(robot_class.class_name, fg='green')}")
    click.echo(f"  Description: {click.style(robot_class.description or 'N/A', fg='yellow')}")


@cli.command("delete")
@click.argument("name")
@coro
async def delete_robot_class(name: str) -> None:
    """Deletes a robot class."""
    async with RobotClassClient() as client:
        await client.delete_robot_class(name)
    click.echo(f"Robot class deleted: {click.style(name, fg='red')}")


@cli.group()
def metadata() -> None:
    """Handle the robot class metadata."""
    pass


@metadata.command("update")
@click.argument("name")
@click.argument("json_path", type=click.Path(exists=True))
@coro
async def update_metadata(name: str, json_path: str) -> None:
    """Updates the metadata of a robot class."""
    with open(json_path, "r", encoding="utf-8") as f:
        raw_metadata = json.load(f)
    metadata = RobotURDFMetadataInputStrict.model_validate(raw_metadata)
    async with RobotClassClient() as client:
        robot_class = await client.update_robot_class(name, new_metadata=metadata)
    click.echo("Robot class metadata updated:")
    click.echo(f"  ID: {click.style(robot_class.id, fg='blue')}")
    click.echo(f"  Name: {click.style(robot_class.class_name, fg='green')}")


@metadata.command("get")
@click.argument("name")
@click.option("--json-path", type=click.Path(exists=False))
@coro
async def get_metadata(name: str, json_path: str | None = None) -> None:
    """Gets the metadata of a robot class."""
    async with RobotClassClient() as client:
        robot_class = await client.get_robot_class(name)
    metadata = robot_class.metadata
    if metadata is None:
        click.echo(click.style("No metadata found", fg="red"))
        return
    if json_path is None:
        click.echo(metadata.model_dump_json(indent=2))
    else:
        with open(json_path, "w", encoding="utf-8") as f:
            json.dump(metadata.model_dump(), f)


@cli.group()
def urdf() -> None:
    """Handle the robot class URDF."""
    pass


@urdf.command("upload")
@click.argument("class_name")
@click.argument("urdf_file")
@coro
async def upload_urdf(class_name: str, urdf_file: str) -> None:
    """Uploads a URDF file to a robot class."""
    async with RobotClassClient() as client:
        response = await client.upload_robot_class_urdf(class_name, urdf_file)
    click.echo("URDF uploaded:")
    click.echo(f"  Filename: {click.style(response.filename, fg='green')}")


@urdf.command("download")
@click.argument("class_name")
@click.option("--cache", is_flag=True, default=False)
@coro
async def download_urdf(class_name: str, cache: bool) -> None:
    """Downloads a URDF file from a robot class."""
    async with RobotClassClient() as client:
        urdf_file = await client.download_and_extract_urdf(class_name, cache=cache)
    click.echo(f"URDF downloaded: {click.style(urdf_file, fg='green')}")


@urdf.command("pybullet")
@click.argument("class_name")
@click.option("--no-cache", is_flag=True, default=False)
@click.option("--hide-gui", is_flag=True, default=False)
@click.option("--hide-origin", is_flag=True, default=False)
@click.option("--see-thru", is_flag=True, default=False)
@click.option("--show-collision", is_flag=True, default=False)
@click.option("--show-inertia", is_flag=True, default=False)
@click.option("--fixed-base", is_flag=True, default=False)
@click.option("--no-merge", is_flag=True, default=False)
@click.option("--dt", type=float, default=0.01)
@click.option("--start-height", type=float, default=0.0)
@click.option("--cycle-duration", type=float, default=2.0)
@coro
async def run_pybullet(
    class_name: str,
    no_cache: bool,
    hide_gui: bool,
    hide_origin: bool,
    see_thru: bool,
    show_collision: bool,
    show_inertia: bool,
    fixed_base: bool,
    no_merge: bool,
    dt: float,
    start_height: float,
    cycle_duration: float,
) -> None:
    """Shows the URDF file for a robot class in PyBullet."""
    try:
        import pybullet as p
    except ImportError:
        click.echo(click.style("PyBullet is not installed; install it with `pip install pybullet`", fg="red"))
        return
    async with RobotClassClient() as client:
        urdf_base = await client.download_and_extract_urdf(class_name, cache=not no_cache)
    try:
        urdf_path = next(urdf_base.glob("*.urdf"))
    except StopIteration:
        click.echo(click.style(f"No URDF file found in {urdf_base}", fg="red"))
        return

    # Connect to PyBullet.
    p.connect(p.GUI)
    p.setGravity(0, 0, -9.81)
    p.setRealTimeSimulation(0)

    # Create floor plane
    floor = p.createCollisionShape(p.GEOM_PLANE)
    p.createMultiBody(0, floor)

    # Turn off panels.
    if hide_gui:
        p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
    p.configureDebugVisualizer(p.COV_ENABLE_SEGMENTATION_MARK_PREVIEW, 0)
    p.configureDebugVisualizer(p.COV_ENABLE_DEPTH_BUFFER_PREVIEW, 0)
    p.configureDebugVisualizer(p.COV_ENABLE_RGB_BUFFER_PREVIEW, 0)

    # Enable mouse picking.
    p.configureDebugVisualizer(p.COV_ENABLE_MOUSE_PICKING, 1)

    # Load the robot URDF.
    start_position = [0.0, 0.0, start_height]
    start_orientation = p.getQuaternionFromEuler([0.0, 0.0, 0.0])
    flags = p.URDF_USE_INERTIA_FROM_FILE
    if not no_merge:
        flags |= p.URDF_MERGE_FIXED_LINKS

    robot = p.loadURDF(
        str(urdf_path.resolve().absolute()),
        start_position,
        start_orientation,
        flags=flags,
        useFixedBase=fixed_base,
    )

    # Display collision meshes as separate object.
    if show_collision:
        collision_flags = p.URDF_USE_INERTIA_FROM_FILE | p.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
        collision = p.loadURDF(
            str(urdf_path.resolve().absolute()),
            start_position,
            start_orientation,
            flags=collision_flags,
            useFixedBase=0,
        )

        # Make collision shapes semi-transparent.
        joint_ids = [i for i in range(p.getNumJoints(collision))] + [-1]
        for i in joint_ids:
            p.changeVisualShape(collision, i, rgbaColor=[1, 0, 0, 0.5])

    # Initializes physics parameters.
    p.changeDynamics(floor, -1, lateralFriction=1, spinningFriction=-1, rollingFriction=-1)
    p.setPhysicsEngineParameter(fixedTimeStep=dt, maxNumCmdPer1ms=1000)

    # Shows the origin of the robot.
    if not hide_origin:
        p.addUserDebugLine([0, 0, 0], [0.1, 0, 0], [1, 0, 0], parentObjectUniqueId=robot, parentLinkIndex=-1)
        p.addUserDebugLine([0, 0, 0], [0, 0.1, 0], [0, 1, 0], parentObjectUniqueId=robot, parentLinkIndex=-1)
        p.addUserDebugLine([0, 0, 0], [0, 0, 0.1], [0, 0, 1], parentObjectUniqueId=robot, parentLinkIndex=-1)

    # Make the robot see-through.
    joint_ids = [i for i in range(p.getNumJoints(robot))] + [-1]
    if see_thru:
        shape_data = p.getVisualShapeData(robot)
        for i in joint_ids:
            prev_color = shape_data[i][-1]
            p.changeVisualShape(robot, i, rgbaColor=prev_color[:3] + (0.9,))

    def draw_box(pt: Sequence[Sequence[float]], color: tuple[float, float, float], obj_id: int, link_id: int) -> None:
        """Draw a box in PyBullet debug visualization.

        Args:
            pt: List of 8 points defining box vertices, each point is [x,y,z]
            color: RGB color tuple for the box lines
            obj_id: PyBullet object ID to attach box to
            link_id: Link ID on the object to attach box to
        """
        assert len(pt) == 8
        assert all(len(p) == 3 for p in pt)

        p.addUserDebugLine(pt[0], pt[1], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[1], pt[3], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[3], pt[2], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[2], pt[0], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)

        p.addUserDebugLine(pt[0], pt[4], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[1], pt[5], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[2], pt[6], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[3], pt[7], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)

        p.addUserDebugLine(pt[4 + 0], pt[4 + 1], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[4 + 1], pt[4 + 3], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[4 + 3], pt[4 + 2], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[4 + 2], pt[4 + 0], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)

    # Shows bounding boxes around each part of the robot representing the inertia frame.
    if show_inertia:
        for i in joint_ids:
            dynamics_info = p.getDynamicsInfo(robot, i)
            mass = dynamics_info[0]
            if mass <= 0:
                continue
            inertia = dynamics_info[2]

            # Calculate box dimensions.
            ixx, iyy, izz = inertia[0], inertia[1], inertia[2]
            box_scale_x = math.sqrt(6 * (iyy + izz - ixx) / mass) / 2
            box_scale_y = math.sqrt(6 * (ixx + izz - iyy) / mass) / 2
            box_scale_z = math.sqrt(6 * (ixx + iyy - izz) / mass) / 2
            half_extents = [box_scale_x, box_scale_y, box_scale_z]

            # Create box vertices in local inertia frame
            pt = [
                [half_extents[0], half_extents[1], half_extents[2]],
                [-half_extents[0], half_extents[1], half_extents[2]],
                [half_extents[0], -half_extents[1], half_extents[2]],
                [-half_extents[0], -half_extents[1], half_extents[2]],
                [half_extents[0], half_extents[1], -half_extents[2]],
                [-half_extents[0], half_extents[1], -half_extents[2]],
                [half_extents[0], -half_extents[1], -half_extents[2]],
                [-half_extents[0], -half_extents[1], -half_extents[2]],
            ]

            draw_box(pt, (1, 0, 0), robot, i)

    # Show joint controller.
    joints: dict[str, int] = {}
    controls: dict[str, float] = {}
    for i in range(p.getNumJoints(robot)):
        joint_info = p.getJointInfo(robot, i)
        name = joint_info[1].decode("utf-8")
        joint_type = joint_info[2]
        joints[name] = i
        if joint_type == p.JOINT_PRISMATIC:
            joint_min, joint_max = joint_info[8:10]
            controls[name] = p.addUserDebugParameter(name, joint_min, joint_max, 0.0)
        elif joint_type == p.JOINT_REVOLUTE:
            joint_min, joint_max = joint_info[8:10]
            controls[name] = p.addUserDebugParameter(name, joint_min, joint_max, 0.0)

    def reset_joints_to_zero(robot: int, joints: dict[str, int]) -> None:
        for joint_id in joints.values():
            joint_info = p.getJointInfo(robot, joint_id)
            joint_min, joint_max = joint_info[8:10]
            zero_position = (joint_min + joint_max) / 2
            p.setJointMotorControl2(robot, joint_id, p.POSITION_CONTROL, zero_position)

    def reset_camera(position: int) -> None:
        height = start_height if fixed_base else 0
        camera_positions = {
            1: (2.0, 0, -30, [0, 0, height]),  # Default view
            2: (2.0, 90, -30, [0, 0, height]),  # Side view
            3: (2.0, 180, -30, [0, 0, height]),  # Back view
            4: (2.0, 270, -30, [0, 0, height]),  # Other side view
            5: (2.0, 0, 0, [0, 0, height]),  # Front level view
            6: (2.0, 0, -80, [0, 0, height]),  # Top-down view
            7: (1.5, 45, -45, [0, 0, height]),  # Closer angled view
            8: (3.0, 30, -30, [0, 0, height]),  # Further angled view
            9: (2.0, 0, 30, [0, 0, height]),  # Low angle view
        }

        if position in camera_positions:
            distance, yaw, pitch, target = camera_positions[position]
            p.resetDebugVisualizerCamera(
                cameraDistance=distance,
                cameraYaw=yaw,
                cameraPitch=pitch,
                cameraTargetPosition=target,
            )

    # Run the simulation until the user closes the window.
    last_time = time.time()
    prev_control_values = {k: 0.0 for k in controls}
    cycle_joints = False
    cycle_start_time = 0.0

    while p.isConnected():
        # Reset the simulation if "r" was pressed.
        keys = p.getKeyboardEvents()
        if ord("r") in keys and keys[ord("r")] & p.KEY_WAS_TRIGGERED:
            p.resetBasePositionAndOrientation(robot, start_position, start_orientation)
            p.setJointMotorControlArray(
                robot,
                range(p.getNumJoints(robot)),
                p.POSITION_CONTROL,
                targetPositions=[0] * p.getNumJoints(robot),
            )

        # Reset joints to zero position if "z" was pressed
        if ord("z") in keys and keys[ord("z")] & p.KEY_WAS_TRIGGERED:
            reset_joints_to_zero(robot, joints)
            cycle_joints = False  # Stop joint cycling if it was active

        # Reset camera if number keys 1-9 are pressed
        for i in range(1, 10):
            if ord(str(i)) in keys and keys[ord(str(i))] & p.KEY_WAS_TRIGGERED:
                reset_camera(i)

        # Start/stop joint cycling if "c" was pressed
        if ord("c") in keys and keys[ord("c")] & p.KEY_WAS_TRIGGERED:
            cycle_joints = not cycle_joints
            if cycle_joints:
                cycle_start_time = time.time()
            else:
                # When stopping joint cycling, set joints to their current positions
                for k, v in controls.items():
                    current_position = p.getJointState(robot, joints[k])[0]
                    p.setJointMotorControl2(robot, joints[k], p.POSITION_CONTROL, current_position)

        # Set joint positions.
        if cycle_joints:
            elapsed_time = time.time() - cycle_start_time
            cycle_progress = (elapsed_time % cycle_duration) / cycle_duration
            for k, v in controls.items():
                joint_info = p.getJointInfo(robot, joints[k])
                joint_min, joint_max = joint_info[8:10]
                target_position = joint_min + (joint_max - joint_min) * math.sin(cycle_progress * math.pi)
                p.setJointMotorControl2(robot, joints[k], p.POSITION_CONTROL, target_position)
        else:
            for k, v in controls.items():
                try:
                    target_position = p.readUserDebugParameter(v)
                    if target_position != prev_control_values[k]:
                        prev_control_values[k] = target_position
                        p.setJointMotorControl2(robot, joints[k], p.POSITION_CONTROL, target_position)
                except p.error:
                    logger.debug("Failed to set joint %s", k)
                    pass

        # Step simulation.
        p.stepSimulation()
        cur_time = time.time()
        time.sleep(max(0, dt - (cur_time - last_time)))
        last_time = cur_time


@urdf.command("mujoco")
@click.argument("class_name")
@click.option("--no-cache", is_flag=True, default=False)
@coro
async def run_mujoco(class_name: str, no_cache: bool) -> None:
    """Shows the URDF file for a robot class in Mujoco.

    This command downloads and extracts the robot class URDF folder,
    searches for an MJCF file (unless --mjcf-path is provided), and then
    launches the Mujoco viewer using the provided MJCF file.
    """
    try:
        import mujoco.viewer
    except ImportError:
        click.echo(click.style("Mujoco is required; install with `pip install mujoco`", fg="red"))
        return

    async with RobotClassClient() as client:
        extracted_folder = await client.download_and_extract_urdf(class_name, cache=not no_cache)

    try:
        mjcf_file = next(
            itertools.chain(
                extracted_folder.glob("*.scene.mjcf"),
                extracted_folder.glob("*.mjcf"),
                extracted_folder.glob("*.xml"),
            )
        )
    except StopIteration:
        click.echo(click.style(f"No MJCF file found in {extracted_folder}", fg="red"))
        return

    mjcf_path_str = str(mjcf_file.resolve())
    click.echo(f"Launching Mujoco viewer with: {click.style(mjcf_path_str, fg='green')}")
    mujoco.viewer.launch_from_path(mjcf_path_str)


if __name__ == "__main__":
    cli()



"""Defines the common interface for the K-Scale Python API."""

__version__ = "0.3.11"

from pathlib import Path

from kscale.web.clients.client import WWWClient as K

ROOT_DIR = Path(__file__).parent



"""Utility functions for file checksums."""

import hashlib
from pathlib import Path
from typing import Tuple

CHUNK_SIZE = 8192


async def calculate_sha256(file_path: str | Path) -> Tuple[str, int]:
    """Calculate SHA256 checksum and size of a file.

    Args:
        file_path: Path to the file

    Returns:
        Tuple of (checksum hex string, file size in bytes)
    """
    sha256_hash = hashlib.sha256()
    file_size = 0

    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(CHUNK_SIZE), b""):
            sha256_hash.update(chunk)
            file_size += len(chunk)

    return sha256_hash.hexdigest(), file_size


class FileChecksum:
    """Helper class for handling file checksums."""

    @staticmethod
    async def calculate(file_path: str | Path) -> Tuple[str, int]:
        """Calculate SHA256 checksum and size of a file."""
        return await calculate_sha256(file_path)

    @staticmethod
    def update_hash(hash_obj: "hashlib._Hash", chunk: bytes) -> None:
        """Update a hash object with new data."""
        hash_obj.update(chunk)



"""Defines a base client for the K-Scale WWW API client."""

import asyncio
import json
import logging
import os
import secrets
import sys
import time
import webbrowser
from types import TracebackType
from typing import Any, Self, Type
from urllib.parse import urljoin

import aiohttp
import httpx
from aiohttp import web
from async_lru import alru_cache
from jwt import ExpiredSignatureError, PyJWKClient, decode as jwt_decode
from pydantic import BaseModel
from yarl import URL

from kscale.web.gen.api import OICDInfo
from kscale.web.utils import DEFAULT_UPLOAD_TIMEOUT, get_api_root, get_auth_dir

logger = logging.getLogger(__name__)

# This port matches the available port for the OAuth callback.
OAUTH_PORT = 16821

# This is the name of the API key header for the K-Scale WWW API.
HEADER_NAME = "x-kscale-api-key"


def verbose_error() -> bool:
    return os.environ.get("KSCALE_VERBOSE_ERROR", "0") == "1"


class OAuthCallback:
    def __init__(self) -> None:
        self.token_type: str | None = None
        self.access_token: str | None = None
        self.id_token: str | None = None
        self.state: str | None = None
        self.expires_in: str | None = None
        self.app = web.Application()
        self.app.router.add_get("/token", self.handle_token)
        self.app.router.add_get("/callback", self.handle_callback)

    async def handle_token(self, request: web.Request) -> web.Response:
        """Handle the token extraction."""
        self.token_type = request.query.get("token_type")
        self.access_token = request.query.get("access_token")
        self.id_token = request.query.get("id_token")
        self.state = request.query.get("state")
        self.expires_in = request.query.get("expires_in")
        return web.Response(text="OK")

    async def handle_callback(self, request: web.Request) -> web.Response:
        """Handle the OAuth callback with token in URL fragment."""
        return web.Response(
            text="""
                <!DOCTYPE html>
                <html lang="en">
                <head>
                    <meta charset="UTF-8">
                    <meta name="viewport" content="width=device-width, initial-scale=1.0">
                    <title>Authentication successful</title>
                    <style>
                        body {
                            font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
                            display: flex;
                            justify-content: center;
                            align-items: center;
                            min-height: 100vh;
                            margin: 0;
                            background: #f5f5f5;
                            color: #333;
                        }
                        .container {
                            background: white;
                            padding: 2rem;
                            border-radius: 8px;
                            box-shadow: 0 2px 4px rgba(0,0,0,0.1);
                            max-width: 600px;
                            width: 90%;
                        }
                        h1 {
                            color: #2c3e50;
                            margin-bottom: 1rem;
                        }
                        .token-info {
                            background: #f8f9fa;
                            border: 1px solid #dee2e6;
                            border-radius: 4px;
                            padding: 1rem;
                            margin: 1rem 0;
                            word-break: break-all;
                        }
                        .token-label {
                            font-weight: bold;
                            color: #6c757d;
                            margin-bottom: 0.5rem;
                        }
                        .success-icon {
                            color: #28a745;
                            font-size: 48px;
                            margin-bottom: 1rem;
                        }
                    </style>
                </head>
                <body>
                    <div class="container">
                        <div class="success-icon">✓</div>
                        <h1>Authentication successful!</h1>
                        <p>Your authentication tokens are shown below. You can now close this window.</p>

                        <div class="token-info">
                            <div class="token-label">Access Token:</div>
                            <div id="accessTokenDisplay"></div>
                        </div>

                        <div class="token-info">
                            <div class="token-label">ID Token:</div>
                            <div id="idTokenDisplay"></div>
                        </div>
                    </div>

                    <script>
                        const params = new URLSearchParams(window.location.hash.substring(1));
                        const tokenType = params.get('token_type');
                        const accessToken = params.get('access_token');
                        const idToken = params.get('id_token');
                        const state = params.get('state');
                        const expiresIn = params.get('expires_in');

                        // Display tokens
                        document.getElementById('accessTokenDisplay').textContent = accessToken || 'Not provided';
                        document.getElementById('idTokenDisplay').textContent = idToken || 'Not provided';

                        if (accessToken) {
                            const tokenUrl = new URL(window.location.href);
                            tokenUrl.pathname = '/token';
                            tokenUrl.searchParams.set('access_token', accessToken);
                            tokenUrl.searchParams.set('token_type', tokenType);
                            tokenUrl.searchParams.set('id_token', idToken);
                            tokenUrl.searchParams.set('state', state);
                            tokenUrl.searchParams.set('expires_in', expiresIn);
                            fetch(tokenUrl.toString());
                        }
                    </script>
                </body>
                </html>
            """,
            content_type="text/html",
        )


class BaseClient:
    def __init__(
        self,
        base_url: str | None = None,
        upload_timeout: float = DEFAULT_UPLOAD_TIMEOUT,
        use_cache: bool = True,
    ) -> None:
        self.base_url = get_api_root() if base_url is None else base_url
        self.upload_timeout = upload_timeout
        self.use_cache = use_cache
        self._client: httpx.AsyncClient | None = None
        self._client_no_auth: httpx.AsyncClient | None = None

    @alru_cache
    async def _get_oicd_info(self) -> OICDInfo:
        cache_path = get_auth_dir() / "oicd_info.json"
        if self.use_cache and cache_path.exists():
            with open(cache_path, "r") as f:
                return OICDInfo(**json.load(f))
        data = await self._request("GET", "/auth/oicd", auth=False)
        if self.use_cache:
            cache_path.parent.mkdir(parents=True, exist_ok=True)
            with open(cache_path, "w") as f:
                json.dump(data, f)
        return OICDInfo(**data)

    @alru_cache
    async def _get_oicd_metadata(self) -> dict:
        """Returns the OpenID Connect server configuration.

        Returns:
            The OpenID Connect server configuration.
        """
        cache_path = get_auth_dir() / "oicd_metadata.json"
        if self.use_cache and cache_path.exists():
            with open(cache_path, "r") as f:
                return json.load(f)
        oicd_info = await self._get_oicd_info()
        oicd_config_url = f"{oicd_info.authority}/.well-known/openid-configuration"
        async with aiohttp.ClientSession() as session:
            async with session.get(oicd_config_url) as response:
                metadata = await response.json()
        if self.use_cache:
            cache_path.parent.mkdir(parents=True, exist_ok=True)
            with open(cache_path, "w") as f:
                json.dump(metadata, f, indent=2)
            logger.info("Cached OpenID Connect metadata to %s", cache_path)
        return metadata

    async def _get_bearer_token(self) -> str:
        """Get a bearer token using the OAuth2 implicit flow.

        Returns:
            A bearer token to use with the K-Scale WWW API.
        """
        # Check if we are in a headless environment.
        error_message = (
            "Cannot perform browser-based authentication in a headless environment. "
            "Please use 'kscale user key' to generate an API key locally and set "
            "the KSCALE_API_KEY environment variable instead."
        )
        try:
            if not webbrowser.get().name != "null":
                raise RuntimeError(error_message)
        except webbrowser.Error:
            raise RuntimeError(error_message)

        oicd_info = await self._get_oicd_info()
        metadata = await self._get_oicd_metadata()
        auth_endpoint = metadata["authorization_endpoint"]

        # Use the cached state and nonce if available, otherwise generate.
        state_file = get_auth_dir() / "oauth_state.json"
        state: str | None = None
        nonce: str | None = None
        if state_file.exists():
            with open(state_file, "r") as f:
                state_data = json.load(f)
                state = state_data.get("state")
                nonce = state_data.get("nonce")
        if state is None:
            state = secrets.token_urlsafe(32)
        if nonce is None:
            nonce = secrets.token_urlsafe(32)

        # Change /oauth2/authorize to /login to use the login endpoint.
        auth_endpoint = auth_endpoint.replace("/oauth2/authorize", "/login")

        auth_url = str(
            URL(auth_endpoint).with_query(
                {
                    "response_type": "token",
                    "redirect_uri": f"http://localhost:{OAUTH_PORT}/callback",
                    "state": state,
                    "nonce": nonce,
                    "scope": "openid profile email",
                    "client_id": oicd_info.client_id,
                }
            )
        )

        # Start local server to receive callback
        callback_handler = OAuthCallback()
        runner = web.AppRunner(callback_handler.app)
        await runner.setup()
        site = web.TCPSite(runner, "localhost", OAUTH_PORT)

        try:
            await site.start()
        except OSError as e:
            raise OSError(
                f"The command line interface requires access to local port {OAUTH_PORT} in order to authenticate with "
                "OpenID Connect. Please ensure that no other application is using this port."
            ) from e

        # Open browser for user authentication
        webbrowser.open(auth_url)

        # Wait for the callback with timeout
        try:
            start_time = time.time()
            while callback_handler.access_token is None:
                if time.time() - start_time > 30:
                    raise TimeoutError("Authentication timed out after 30 seconds")
                await asyncio.sleep(0.1)

            # Save the state and nonce to the cache.
            state = callback_handler.state
            state_file.parent.mkdir(parents=True, exist_ok=True)
            state_file.write_text(json.dumps({"state": state, "nonce": nonce}))

            return callback_handler.access_token
        finally:
            await runner.cleanup()

    @alru_cache
    async def _get_jwk_client(self) -> PyJWKClient:
        """Returns a JWK client for the OpenID Connect server."""
        oicd_info = await self._get_oicd_info()
        jwks_uri = f"{oicd_info.authority}/.well-known/jwks.json"
        return PyJWKClient(uri=jwks_uri)

    async def _is_token_expired(self, token: str) -> bool:
        """Check if a token is expired."""
        jwk_client = await self._get_jwk_client()
        signing_key = jwk_client.get_signing_key_from_jwt(token)

        try:
            claims = jwt_decode(
                token,
                signing_key.key,
                algorithms=["RS256"],
                options={"verify_aud": False},
            )
        except ExpiredSignatureError:
            return True

        return claims["exp"] < time.time()

    @alru_cache
    async def get_bearer_token(self) -> str:
        """Get a bearer token from OpenID Connect.

        Returns:
            A bearer token to use with the K-Scale WWW API.
        """
        cache_path = get_auth_dir() / "bearer_token.txt"
        if self.use_cache and cache_path.exists():
            token = cache_path.read_text()
            if not await self._is_token_expired(token):
                return token
        token = await self._get_bearer_token()
        if self.use_cache:
            cache_path.write_text(token)
            cache_path.chmod(0o600)
        return token

    async def get_client(self, *, auth: bool = True) -> httpx.AsyncClient:
        client = self._client if auth else self._client_no_auth
        if client is None:
            headers: dict[str, str] = {}
            if auth:
                if "KSCALE_API_KEY" in os.environ:
                    headers[HEADER_NAME] = os.environ["KSCALE_API_KEY"]
                else:
                    headers["Authorization"] = f"Bearer {await self.get_bearer_token()}"

            client = httpx.AsyncClient(
                base_url=self.base_url,
                headers=headers,
                timeout=httpx.Timeout(30.0),
            )
            if auth:
                self._client = client
            else:
                self._client_no_auth = client
        return client

    async def _request(
        self,
        method: str,
        endpoint: str,
        *,
        auth: bool = True,
        params: dict[str, Any] | None = None,
        data: BaseModel | dict[str, Any] | None = None,
        files: dict[str, Any] | None = None,
    ) -> dict[str, Any]:
        url = urljoin(self.base_url, endpoint)
        kwargs: dict[str, Any] = {}
        if params is not None:
            kwargs["params"] = params
        if data is not None:
            if isinstance(data, BaseModel):
                kwargs["json"] = data.model_dump(exclude_unset=True)
            else:
                kwargs["json"] = data
        if files:
            kwargs["files"] = files

        client = await self.get_client(auth=auth)
        response = await client.request(method, url, **kwargs)

        if response.is_error:
            logger.error("Got %d error K-Scale: %s", response.status_code, response.text)
            if verbose_error():
                response.raise_for_status()
            else:
                logger.error("Use KSCALE_VERBOSE_ERROR=1 to see the full error message")
                logger.error("If this persists, please create an issue here: https://github.com/kscalelabs/kscale")
                sys.exit(1)

        return response.json()

    async def close(self) -> None:
        if self._client is not None:
            await self._client.aclose()
            self._client = None

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()



"""Defines the client for interacting with the K-Scale robot class endpoints."""

import hashlib
import json
import logging
import tarfile
from pathlib import Path
from typing import Any

import httpx

from kscale.web.clients.base import BaseClient
from kscale.web.gen.api import (
    RobotClass,
    RobotDownloadURDFResponse,
    RobotUploadURDFResponse,
    RobotURDFMetadataInput,
)
from kscale.web.utils import get_robots_dir, should_refresh_file

logger = logging.getLogger(__name__)

UPLOAD_TIMEOUT = 300.0
DOWNLOAD_TIMEOUT = 60.0

INFO_FILE_NAME = ".info.json"


class RobotClassClient(BaseClient):
    async def get_robot_classes(self) -> list[RobotClass]:
        data = await self._request(
            "GET",
            "/robots/",
            auth=True,
        )
        return [RobotClass.model_validate(item) for item in data]

    async def get_robot_class(self, class_name: str) -> RobotClass:
        data = await self._request(
            "GET",
            f"/robots/name/{class_name}",
            auth=True,
        )
        return RobotClass.model_validate(data)

    async def create_robot_class(self, class_name: str, description: str | None = None) -> RobotClass:
        data = {}
        if description is not None:
            data["description"] = description
        data = await self._request(
            "PUT",
            f"/robots/{class_name}",
            data=data,
            auth=True,
        )
        return RobotClass.model_validate(data)

    async def update_robot_class(
        self,
        class_name: str,
        new_class_name: str | None = None,
        new_description: str | None = None,
        new_metadata: RobotURDFMetadataInput | None = None,
    ) -> RobotClass:
        data: dict[str, Any] = {}
        if new_class_name is not None:
            data["new_class_name"] = new_class_name
        if new_description is not None:
            data["new_description"] = new_description
        if new_metadata is not None:
            data["new_metadata"] = new_metadata.model_dump()
        if not data:
            raise ValueError("No parameters to update")
        data = await self._request(
            "POST",
            f"/robots/{class_name}",
            data=data,
            auth=True,
        )
        return RobotClass.model_validate(data)

    async def delete_robot_class(self, class_name: str) -> None:
        await self._request("DELETE", f"/robots/{class_name}", auth=True)

    async def upload_robot_class_urdf(self, class_name: str, urdf_file: str | Path) -> RobotUploadURDFResponse:
        if not (urdf_file := Path(urdf_file)).exists():
            raise FileNotFoundError(f"URDF file not found: {urdf_file}")

        # Gets the content type from the file extension.
        ext = urdf_file.suffix.lower()
        match ext:
            case ".tgz":
                content_type = "application/x-compressed-tar"
            case _:
                raise ValueError(f"Unsupported file extension: {ext}")

        data = await self._request(
            "PUT",
            f"/robots/urdf/{class_name}",
            data={"filename": urdf_file.name, "content_type": content_type},
            auth=True,
        )
        response = RobotUploadURDFResponse.model_validate(data)
        async with httpx.AsyncClient(
            timeout=httpx.Timeout(UPLOAD_TIMEOUT),
        ) as client:
            async with client.stream(
                "PUT",
                response.url,
                content=urdf_file.read_bytes(),
                headers={"Content-Type": response.content_type},
            ) as r:
                r.raise_for_status()
        return response

    async def download_compressed_urdf(self, class_name: str, *, cache: bool = True) -> Path:
        cache_path = get_robots_dir() / class_name / "robot.tgz"
        if cache and cache_path.exists() and not should_refresh_file(cache_path):
            return cache_path
        data = await self._request("GET", f"/robots/urdf/{class_name}", auth=True)
        response = RobotDownloadURDFResponse.model_validate(data)
        expected_hash = response.md5_hash
        cache_path.parent.mkdir(parents=True, exist_ok=True)

        # Checks the md5 hash of the file.
        cache_path_info = cache_path.parent / INFO_FILE_NAME
        if cache_path_info.exists():
            with open(cache_path_info, "r") as f:
                info = json.load(f)
                if info["md5_hash"] == expected_hash:
                    cache_path.touch()
                    return cache_path

        logger.info("Downloading URDF file from %s", response.url)
        async with httpx.AsyncClient(
            timeout=httpx.Timeout(DOWNLOAD_TIMEOUT),
        ) as client:
            with open(cache_path, "wb") as file:
                hash_value = hashlib.md5()
                async with client.stream("GET", response.url) as r:
                    r.raise_for_status()
                    async for chunk in r.aiter_bytes():
                        file.write(chunk)
                        hash_value.update(chunk)

        logger.info("Checking MD5 hash of downloaded file")
        hash_value_hex = f'"{hash_value.hexdigest()}"'
        if hash_value_hex != expected_hash:
            raise ValueError(f"MD5 hash mismatch: {hash_value_hex} != {expected_hash}")

        logger.info("Updating downloaded file information")
        info = {"md5_hash": hash_value_hex}
        with open(cache_path_info, "w") as f:
            json.dump(info, f)

        return cache_path

    async def download_and_extract_urdf(self, class_name: str, *, cache: bool = True) -> Path:
        cache_path = await self.download_compressed_urdf(class_name, cache=cache)

        # Reads the MD5 hash from the info file.
        cache_path_info = cache_path.parent / INFO_FILE_NAME
        with open(cache_path_info, "r") as f:
            info = json.load(f)
            expected_hash = info["md5_hash"]

        # Unpacks the file if requested.
        unpack_path = cache_path.parent / "robot"
        unpack_path.mkdir(parents=True, exist_ok=True)
        unpacked_path_info = unpack_path / INFO_FILE_NAME

        # If the file has already been unpacked, return the path.
        if unpacked_path_info.exists():
            with open(unpacked_path_info, "r") as f:
                info = json.load(f)
                if info["md5_hash"] == expected_hash:
                    unpack_path.touch()
                    return unpack_path

        logger.info("Unpacking URDF file")
        with tarfile.open(cache_path, "r:gz") as tar:
            tar.extractall(path=unpack_path)

        logger.info("Updating downloaded file information")
        info = {"md5_hash": expected_hash}
        with open(unpacked_path_info, "w") as f:
            json.dump(info, f)

        return unpack_path



"""Utility functions for interacting with the K-Scale WWW API."""

import functools
import logging
import time
from pathlib import Path

from kscale.conf import Settings

logger = logging.getLogger(__name__)

DEFAULT_UPLOAD_TIMEOUT = 300.0  # 5 minutes


@functools.lru_cache
def get_kscale_dir() -> Path:
    """Returns the cache directory for artifacts."""
    return Path(Settings.load().www.base_dir).expanduser().resolve()


def get_auth_dir() -> Path:
    """Returns the directory for authentication artifacts."""
    return get_kscale_dir() / "auth"


def get_robots_dir() -> Path:
    """Returns the directory for robot artifacts."""
    return get_kscale_dir() / "robots"


def should_refresh_file(file: Path) -> bool:
    """Returns whether the file should be refreshed."""
    return file.exists() and file.stat().st_mtime < time.time() - Settings.load().www.refresh_interval_minutes * 60


@functools.lru_cache
def get_api_root() -> str:
    """Returns the root URL for the K-Scale WWW API."""
    return Settings.load().www.api_root



"""Defines a unified client for the K-Scale WWW API."""

from kscale.web.clients.base import BaseClient
from kscale.web.clients.robot import RobotClient
from kscale.web.clients.robot_class import RobotClassClient
from kscale.web.clients.user import UserClient


class WWWClient(
    RobotClient,
    RobotClassClient,
    UserClient,
    BaseClient,
):
    pass



"""Defines the CLI for getting information about the current user."""

import logging

import click
from tabulate import tabulate

from kscale.utils.cli import coro
from kscale.web.clients.user import UserClient

logger = logging.getLogger(__name__)


@click.group()
def cli() -> None:
    """Get information about the currently-authenticated user."""
    pass


@cli.command()
@coro
async def me() -> None:
    """Get information about the currently-authenticated user."""
    client = UserClient()
    profile = await client.get_profile_info()
    click.echo(
        tabulate(
            [
                ["Email", profile.email],
                ["Email verified", profile.email_verified],
                ["User ID", profile.user.user_id],
                ["Is admin", profile.user.is_admin],
                ["Can upload", profile.user.can_upload],
                ["Can test", profile.user.can_test],
            ],
            headers=["Key", "Value"],
            tablefmt="simple",
        )
    )


@cli.command()
@coro
async def key() -> None:
    """Get an API key for the currently-authenticated user."""
    client = UserClient()
    api_key = await client.get_api_key()
    click.echo(f"API key: {click.style(api_key, fg='green')}")


if __name__ == "__main__":
    cli()



"""Defines utilities for working with asyncio."""

import asyncio
import textwrap
from functools import wraps
from typing import Any, Callable, Coroutine, ParamSpec, TypeVar

import click

T = TypeVar("T")
P = ParamSpec("P")


def coro(f: Callable[P, Coroutine[Any, Any, T]]) -> Callable[P, T]:
    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return asyncio.run(f(*args, **kwargs))

    return wrapper


def recursive_help(cmd: click.Command, parent: click.Context | None = None, indent: int = 0) -> str:
    ctx = click.core.Context(cmd, info_name=cmd.name, parent=parent)
    help_text = cmd.get_help(ctx)
    commands = getattr(cmd, "commands", {})
    for sub in commands.values():
        help_text += recursive_help(sub, ctx, indent + 2)
    return textwrap.indent(help_text, " " * indent)



"""Defines the top-level KOL CLI."""

import logging

import click
import colorlogging

from kscale.utils.cli import recursive_help
from kscale.web.cli.robot import cli as robot_cli
from kscale.web.cli.robot_class import cli as robot_class_cli
from kscale.web.cli.user import cli as user_cli


@click.group()
def cli() -> None:
    """Command line interface for interacting with the K-Scale web API."""
    colorlogging.configure()

    # Suppress aiohttp access logging
    logging.getLogger("httpx").setLevel(logging.WARNING)
    logging.getLogger("aiohttp.access").setLevel(logging.WARNING)


cli.add_command(user_cli, "user")
cli.add_command(robot_class_cli, "robots")
cli.add_command(robot_cli, "robot")

if __name__ == "__main__":
    # python -m kscale.cli
    print(recursive_help(cli))



"""Defines the bot environment settings."""

import functools
import os
import warnings
from dataclasses import dataclass, field
from pathlib import Path

from omegaconf import II, OmegaConf

# This is the public API endpoint for the K-Scale WWW API.
DEFAULT_API_ROOT = "https://api.kscale.dev"

SETTINGS_FILE_NAME = "settings.yaml"


def get_path() -> Path:
    if "KSCALE_CONFIG_DIR" in os.environ:
        return Path(os.environ["KSCALE_CONFIG_DIR"]).expanduser().resolve()
    return Path("~/.kscale/").expanduser().resolve()


@dataclass
class WWWSettings:
    api_root: str = field(default=DEFAULT_API_ROOT)
    base_dir: str = field(default=II("oc.env:KSCALE_DIR,'~/.kscale/'"))
    refresh_interval_minutes: int = field(default=60 * 24)


@dataclass
class Settings:
    www: WWWSettings = field(default_factory=WWWSettings)

    def save(self) -> None:
        (dir_path := get_path()).mkdir(parents=True, exist_ok=True)
        with open(dir_path / "settings.yaml", "w") as f:
            OmegaConf.save(config=self, f=f)

    @functools.lru_cache
    @staticmethod
    def load() -> "Settings":
        config = OmegaConf.structured(Settings)
        if not (dir_path := get_path()).exists():
            warnings.warn(f"Settings directory does not exist: {dir_path}. Creating it now.")
            dir_path.mkdir(parents=True)
            OmegaConf.save(config, dir_path / SETTINGS_FILE_NAME)
        else:
            try:
                with open(dir_path / SETTINGS_FILE_NAME, "r") as f:
                    raw_settings = OmegaConf.load(f)
                    config = OmegaConf.merge(config, raw_settings)
            except Exception as e:
                warnings.warn(f"Failed to load settings: {e}")
        return config



"""Auto-generated by generate.sh script."""

# generated by datamodel-codegen:
#   filename:  openapi.json
#   timestamp: 2025-01-22T23:12:27+00:00

from __future__ import annotations

from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field


class APIKeyRequest(BaseModel):
    num_hours: Optional[int] = Field(24, title="Num Hours")


class APIKeyResponse(BaseModel):
    api_key: str = Field(..., title="Api Key")


class AddRobotClassRequest(BaseModel):
    description: Optional[str] = Field(None, title="Description")


class AddRobotRequest(BaseModel):
    description: Optional[str] = Field(None, title="Description")
    class_name: str = Field(..., title="Class Name")


class JointMetadataInput(BaseModel):
    id: Optional[int] = Field(None, title="Id")
    kp: Optional[Union[float, str]] = Field(None, title="Kp")
    kd: Optional[Union[float, str]] = Field(None, title="Kd")
    armature: Optional[Union[float, str]] = Field(None, title="Armature")
    friction: Optional[Union[float, str]] = Field(None, title="Friction")
    offset: Optional[Union[float, str]] = Field(None, title="Offset")
    flipped: Optional[bool] = Field(None, title="Flipped")
    actuator_type: Optional[str] = Field(None, title="Actuator Type")
    nn_id: Optional[int] = Field(None, title="Neural Network Id")
    soft_torque_limit: Optional[str] = Field(None, title="Soft Torque Limit")


class JointMetadataOutput(BaseModel):
    id: Optional[int] = Field(None, title="Id")
    kp: Optional[str] = Field(None, title="Kp")
    kd: Optional[str] = Field(None, title="Kd")
    armature: Optional[str] = Field(None, title="Armature")
    friction: Optional[str] = Field(None, title="Friction")
    offset: Optional[str] = Field(None, title="Offset")
    flipped: Optional[bool] = Field(None, title="Flipped")
    actuator_type: Optional[str] = Field(None, title="Actuator Type")
    nn_id: Optional[int] = Field(None, title="Neural Network Id")
    soft_torque_limit: Optional[str] = Field(None, title="Soft Torque Limit")


class OICDInfo(BaseModel):
    authority: str = Field(..., title="Authority")
    client_id: str = Field(..., title="Client Id")


class Robot(BaseModel):
    id: str = Field(..., title="Id")
    robot_name: str = Field(..., title="Robot Name")
    description: str = Field(..., title="Description")
    user_id: str = Field(..., title="User Id")
    class_id: str = Field(..., title="Class Id")


class RobotDownloadURDFResponse(BaseModel):
    url: str = Field(..., title="Url")
    md5_hash: str = Field(..., title="Md5 Hash")


class RobotResponse(BaseModel):
    id: str = Field(..., title="Id")
    robot_name: str = Field(..., title="Robot Name")
    description: str = Field(..., title="Description")
    user_id: str = Field(..., title="User Id")
    class_name: str = Field(..., title="Class Name")


class RobotURDFMetadataInput(BaseModel):
    joint_name_to_metadata: Optional[Dict[str, JointMetadataInput]] = Field(None, title="Joint Name To Metadata")
    control_frequency: Optional[Union[float, str]] = Field(None, title="Control Frequency")


class RobotURDFMetadataOutput(BaseModel):
    joint_name_to_metadata: Optional[Dict[str, JointMetadataOutput]] = Field(None, title="Joint Name To Metadata")
    control_frequency: Optional[str] = Field(None, title="Control Frequency")


class RobotUploadURDFRequest(BaseModel):
    filename: str = Field(..., title="Filename")
    content_type: str = Field(..., title="Content Type")


class RobotUploadURDFResponse(BaseModel):
    url: str = Field(..., title="Url")
    filename: str = Field(..., title="Filename")
    content_type: str = Field(..., title="Content Type")


class UpdateRobotClassRequest(BaseModel):
    new_class_name: Optional[str] = Field(None, title="New Class Name")
    new_description: Optional[str] = Field(None, title="New Description")
    new_metadata: Optional[RobotURDFMetadataInput] = None


class UpdateRobotRequest(BaseModel):
    new_robot_name: Optional[str] = Field(None, title="New Robot Name")
    new_description: Optional[str] = Field(None, title="New Description")


class UserResponse(BaseModel):
    user_id: str = Field(..., title="User Id")
    is_admin: bool = Field(..., title="Is Admin")
    can_upload: bool = Field(..., title="Can Upload")
    can_test: bool = Field(..., title="Can Test")


class ValidationError(BaseModel):
    loc: List[Union[str, int]] = Field(..., title="Location")
    msg: str = Field(..., title="Message")
    type: str = Field(..., title="Error Type")


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title="Detail")


class ProfileResponse(BaseModel):
    email: str = Field(..., title="Email")
    email_verified: bool = Field(..., title="Email Verified")
    user: UserResponse


class RobotClass(BaseModel):
    id: str = Field(..., title="Id")
    class_name: str = Field(..., title="Class Name")
    description: str = Field(..., title="Description")
    user_id: str = Field(..., title="User Id")
    metadata: Optional[RobotURDFMetadataOutput] = None



"""Defines some helper functions for working with artifacts."""

from pathlib import Path

ARTIFACTS_DIR = Path(__file__).parent.resolve()

PLANE_OBJ_PATH = ARTIFACTS_DIR / "plane.obj"
PLANE_URDF_PATH = ARTIFACTS_DIR / "plane.urdf"



"""Defines the client for interacting with the K-Scale authentication endpoints."""

from kscale.web.clients.base import BaseClient
from kscale.web.gen.api import ProfileResponse


class UserClient(BaseClient):
    async def get_profile_info(self) -> ProfileResponse:
        data = await self._request("GET", "/auth/profile", auth=True)
        return ProfileResponse(**data)

    async def get_api_key(self, num_hours: int = 24) -> str:
        data = await self._request("POST", "/auth/key", auth=True, data={"num_hours": num_hours})
        return data["api_key"]



"""Defines the client for interacting with the K-Scale robot endpoints."""

from kscale.web.clients.base import BaseClient
from kscale.web.gen.api import Robot, RobotResponse


class RobotClient(BaseClient):
    async def get_all_robots(self) -> list[Robot]:
        data = await self._request("GET", "/robot/", auth=True)
        return [Robot.model_validate(item) for item in data]

    async def get_user_robots(self, user_id: str = "me") -> list[Robot]:
        data = await self._request("GET", f"/robot/user/{user_id}", auth=True)
        return [Robot.model_validate(item) for item in data]

    async def add_robot(
        self,
        robot_name: str,
        class_name: str,
        description: str | None = None,
    ) -> RobotResponse:
        data = {"class_name": class_name}
        if description is not None:
            data["description"] = description
        response = await self._request(
            "PUT",
            f"/robot/{robot_name}",
            data=data,
            auth=True,
        )
        return RobotResponse.model_validate(response)

    async def get_robot_by_id(self, robot_id: str) -> RobotResponse:
        data = await self._request("GET", f"/robot/id/{robot_id}", auth=True)
        return RobotResponse.model_validate(data)

    async def get_robot_by_name(self, robot_name: str) -> RobotResponse:
        data = await self._request("GET", f"/robot/name/{robot_name}", auth=True)
        return RobotResponse.model_validate(data)



"""Defines the CLI for getting information about robots."""

import click
from tabulate import tabulate

from kscale.utils.cli import coro
from kscale.web.clients.robot import RobotClient


@click.group()
def cli() -> None:
    """Get information about robots."""
    pass


@cli.command()
@coro
async def list() -> None:
    client = RobotClient()
    robots = await client.get_all_robots()
    if robots:
        table_data = [
            [
                click.style(robot.id, fg="blue"),
                click.style(robot.robot_name, fg="green"),
                click.style(robot.class_id, fg="yellow"),
                robot.description or "N/A",
            ]
            for robot in robots
        ]
        click.echo(tabulate(table_data, headers=["ID", "Name", "Class", "Description"], tablefmt="simple"))
    else:
        click.echo(click.style("No robots found", fg="red"))


@cli.command()
@click.option("-u", "--user-id", type=str, default="me")
@coro
async def user(user_id: str = "me") -> None:
    client = RobotClient()
    robots = await client.get_user_robots(user_id)
    if robots:
        table_data = [
            [
                click.style(robot.id, fg="blue"),
                click.style(robot.robot_name, fg="green"),
                click.style(robot.class_id, fg="yellow"),
                robot.description or "N/A",
            ]
            for robot in robots
        ]
        click.echo(tabulate(table_data, headers=["ID", "Name", "Class", "Description"], tablefmt="simple"))
    else:
        click.echo(click.style("No robots found", fg="red"))


@cli.command()
@click.argument("robot_id")
@coro
async def id(robot_id: str) -> None:
    client = RobotClient()
    robot = await client.get_robot_by_id(robot_id)
    click.echo("Robot:")
    click.echo(f"  ID: {click.style(robot.id, fg='blue')}")
    click.echo(f"  Name: {click.style(robot.robot_name, fg='green')}")
    click.echo(f"  Class: {click.style(robot.class_name, fg='yellow')}")
    click.echo(f"  Description: {click.style(robot.description or 'N/A', fg='yellow')}")


@cli.command()
@click.argument("robot_name")
@coro
async def name(robot_name: str) -> None:
    client = RobotClient()
    robot = await client.get_robot_by_name(robot_name)
    click.echo("Robot:")
    click.echo(f"  ID: {click.style(robot.id, fg='blue')}")
    click.echo(f"  Name: {click.style(robot.robot_name, fg='green')}")
    click.echo(f"  Class: {click.style(robot.class_name, fg='yellow')}")
    click.echo(f"  Description: {click.style(robot.description or 'N/A', fg='yellow')}")


@cli.command()
@click.argument("class_name")
@click.argument("name")
@click.option("-c", "--class-name", type=str, required=True)
@click.option("-d", "--description", type=str, default=None)
@coro
async def add(name: str, class_name: str, description: str | None = None) -> None:
    client = RobotClient()
    robot = await client.add_robot(name, class_name, description)
    click.echo("Robot added:")
    click.echo(f"  ID: {click.style(robot.id, fg='blue')}")
    click.echo(f"  Name: {click.style(robot.robot_name, fg='green')}")
    click.echo(f"  Class: {click.style(robot.class_name, fg='yellow')}")
    click.echo(f"  Description: {click.style(robot.description or 'N/A', fg='yellow')}")


if __name__ == "__main__":
    cli()
